@@ -1,6 +1,6 @@
 # -*- encoding: binary -*-
-require File.expand_path('../../../spec_helper', __FILE__)
-require File.expand_path('../shared/extract_range', __FILE__)
+require_relative '../../spec_helper'
+require_relative 'shared/extract_range'
 require 'strscan'
 
 describe "StringScanner#getch" do
